@@ -23,13 +23,13 @@ import Prelude hiding (head, take, takeWhile, dropWhile)
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
 import Data.Conduit
-import qualified Data.Conduit.List as CL
 import Control.Exception (assert)
 import Control.Monad (liftM)
-import Control.Monad.IO.Class (liftIO)
+import Control.Monad.IO.Class (liftIO, MonadIO)
 import qualified System.IO as IO
-import Control.Monad.Trans.Resource (withIO, release)
+import Control.Monad.Trans.Resource (allocate, release)
 import Data.Word (Word8)
+import Data.Monoid (mempty)
 #if CABAL_OS_WINDOWS
 import qualified System.Win32File as F
 #elif NO_HANDLES
@@ -42,17 +42,17 @@ import qualified System.PosixFile as F
 -- While you are not required to call @hClose@ on the resulting handle, you
 -- should do so as early as possible to free scarce resources.
 --
--- Since 0.2.0
-openFile :: ResourceIO m
+-- Since 0.3.0
+openFile :: MonadResource m
          => FilePath
          -> IO.IOMode
-         -> ResourceT m IO.Handle
-openFile fp mode = fmap snd $ withIO (IO.openBinaryFile fp mode) IO.hClose
+         -> m IO.Handle
+openFile fp mode = liftM snd $ allocate (IO.openBinaryFile fp mode) IO.hClose
 
 -- | Stream the contents of a file as binary data.
 --
--- Since 0.2.0
-sourceFile :: ResourceIO m
+-- Since 0.3.0
+sourceFile :: MonadResource m
            => FilePath
            -> Source m S.ByteString
 sourceFile fp =
@@ -68,20 +68,20 @@ sourceFile fp =
 -- function will /not/ automatically close the @Handle@ when processing
 -- completes, since it did not acquire the @Handle@ in the first place.
 --
--- Since 0.2.0
-sourceHandle :: ResourceIO m
+-- Since 0.3.0
+sourceHandle :: MonadIO m
              => IO.Handle
              -> Source m S.ByteString
 sourceHandle h =
     src
   where
-    src = Source pull close
+    src = PipeM pull close
 
     pull = do
         bs <- liftIO (S.hGetSome h 4096)
         if S.null bs
-            then return Closed
-            else return $ Open src bs
+            then return $ Done Nothing ()
+            else return $ HaveOutput src close bs
 
     close = return ()
 
@@ -90,8 +90,8 @@ sourceHandle h =
 -- a 'IO.Handle' (in read mode), so that it can open it only when needed
 -- and close it as soon as possible.
 --
--- Since 0.2.0
-sourceIOHandle :: ResourceIO m
+-- Since 0.3.0
+sourceIOHandle :: MonadResource m
                => IO IO.Handle
                -> Source m S.ByteString
 sourceIOHandle alloc = sourceIO alloc IO.hClose
@@ -104,14 +104,16 @@ sourceIOHandle alloc = sourceIO alloc IO.hClose
 -- | Stream all incoming data to the given 'IO.Handle'. Note that this function
 -- will /not/ automatically close the @Handle@ when processing completes.
 --
--- Since 0.2.0
-sinkHandle :: ResourceIO m
+-- Since 0.3.0
+sinkHandle :: MonadIO m
            => IO.Handle
            -> Sink S.ByteString m ()
 sinkHandle h =
-    SinkData push close
+    NeedInput push close
   where
-    push input = liftIO (S.hPut h input) >> return (Processing push close)
+    push input = PipeM
+        (liftIO (S.hPut h input) >> return (NeedInput push close))
+        (return ())
     close = return ()
 
 -- | An alternative to 'sinkHandle'.
@@ -119,8 +121,8 @@ sinkHandle h =
 -- a 'IO.Handle' (in write mode), so that it can open it only when needed
 -- and close it as soon as possible.
 --
--- Since 0.2.0
-sinkIOHandle :: ResourceIO m
+-- Since 0.3.0
+sinkIOHandle :: MonadResource m
              => IO IO.Handle
              -> Sink S.ByteString m ()
 sinkIOHandle alloc = sinkIO alloc IO.hClose
@@ -130,36 +132,34 @@ sinkIOHandle alloc = sinkIO alloc IO.hClose
 -- | Stream the contents of a file as binary data, starting from a certain
 -- offset and only consuming up to a certain number of bytes.
 --
--- Since 0.2.0
-sourceFileRange :: ResourceIO m
+-- Since 0.3.0
+sourceFileRange :: MonadResource m
                 => FilePath
                 -> Maybe Integer -- ^ Offset
                 -> Maybe Integer -- ^ Maximum count
                 -> Source m S.ByteString
-sourceFileRange fp offset count = Source
-    { sourcePull = do
-        (key, handle) <- withIO (IO.openBinaryFile fp IO.ReadMode) IO.hClose
+sourceFileRange fp offset count = PipeM
+    (do
+        (key, handle) <- allocate (IO.openBinaryFile fp IO.ReadMode) IO.hClose
         case offset of
             Nothing -> return ()
             Just off -> liftIO $ IO.hSeek handle IO.AbsoluteSeek off
         case count of
             Nothing -> pullUnlimited handle key
-            Just c -> pullLimited c handle key
-    , sourceClose = return ()
-    }
+            Just c -> pullLimited c handle key)
+    (return ())
   where
     pullUnlimited handle key = do
         bs <- liftIO $ S.hGetSome handle 4096
         if S.null bs
             then do
                 release key
-                return Closed
+                return $ Done Nothing ()
             else do
-                let src = Source
-                        { sourcePull = pullUnlimited handle key
-                        , sourceClose = release key
-                        }
-                return $ Open src bs
+                let src = PipeM
+                        (pullUnlimited handle key)
+                        (release key)
+                return $ HaveOutput src (release key) bs
 
     pullLimited c0 handle key = do
         let c = fromInteger c0
@@ -169,18 +169,17 @@ sourceFileRange fp offset count = Source
             if S.null bs
                 then do
                     release key
-                    return Closed
+                    return $ Done Nothing ()
                 else do
-                    let src = Source
-                            { sourcePull = pullLimited (toInteger c') handle key
-                            , sourceClose = release key
-                            }
-                    return $ Open src bs
+                    let src = PipeM
+                            (pullLimited (toInteger c') handle key)
+                            (release key)
+                    return $ HaveOutput src (release key) bs
 
 -- | Stream all incoming data to the given file.
 --
--- Since 0.2.0
-sinkFile :: ResourceIO m
+-- Since 0.3.0
+sinkFile :: MonadResource m
          => FilePath
          -> Sink S.ByteString m ()
 sinkFile fp = sinkIOHandle (IO.openBinaryFile fp IO.WriteMode)
@@ -188,8 +187,8 @@ sinkFile fp = sinkIOHandle (IO.openBinaryFile fp IO.WriteMode)
 -- | Stream the contents of the input to a file, and also send it along the
 -- pipeline. Similar in concept to the Unix command @tee@.
 --
--- Since 0.2.0
-conduitFile :: ResourceIO m
+-- Since 0.3.0
+conduitFile :: MonadResource m
             => FilePath
             -> Conduit S.ByteString m S.ByteString
 conduitFile fp = conduitIO
@@ -204,8 +203,8 @@ conduitFile fp = conduitIO
 -- sink. Note that this does /not/ ensure that all of those bytes are in fact
 -- consumed.
 --
--- Since 0.2.0
-isolate :: Resource m
+-- Since 0.3.0
+isolate :: Monad m
         => Int
         -> Conduit S.ByteString m S.ByteString
 isolate count0 = conduitState
@@ -225,78 +224,94 @@ isolate count0 = conduitState
 
 -- | Return the next byte from the stream, if available.
 --
--- Since 0.2.0
-head :: Resource m => Sink S.ByteString m (Maybe Word8)
+-- Since 0.3.0
+head :: Monad m => Sink S.ByteString m (Maybe Word8)
 head =
-    SinkData push close
+    NeedInput push close
   where
     push bs =
         case S.uncons bs of
-            Nothing -> return $ Processing push close
-            Just (w, bs') -> do
+            Nothing -> NeedInput push close
+            Just (w, bs') ->
                 let lo = if S.null bs' then Nothing else Just bs'
-                return $ Done lo (Just w)
+                 in Done lo (Just w)
     close = return Nothing
 
 -- | Return all bytes while the predicate returns @True@.
 --
--- Since 0.2.0
-takeWhile :: Resource m => (Word8 -> Bool) -> Conduit S.ByteString m S.ByteString
+-- Since 0.3.0
+takeWhile :: Monad m => (Word8 -> Bool) -> Conduit S.ByteString m S.ByteString
 takeWhile p =
-    conduit
+    NeedInput push close
   where
-    conduit = Conduit push close
-    push bs = do
-        let (x, y) = S.span p bs
-        return $
-            if S.null y
-                then Producing conduit [x]
-                else Finished (Just y) (if S.null x then [] else [x])
-    close = return []
+    push bs
+        | S.null y =
+            let r = NeedInput push close
+             in if S.null x
+                    then r
+                    else HaveOutput r (return ()) x
+        | otherwise =
+            let f = Done (Just y) ()
+             in if S.null x
+                    then f
+                    else HaveOutput f (return ()) x
+      where
+        (x, y) = S.span p bs
+    close = mempty
 
 -- | Ignore all bytes while the predicate returns @True@.
 --
--- Since 0.2.0
-dropWhile :: Resource m => (Word8 -> Bool) -> Sink S.ByteString m ()
+-- Since 0.3.0
+dropWhile :: Monad m => (Word8 -> Bool) -> Sink S.ByteString m ()
 dropWhile p =
-    SinkData push close
+    NeedInput push close
   where
-    push bs = do
-        let bs' = S.dropWhile p bs
-        return $
-            if S.null bs'
-                then Processing push close
-                else Done (Just bs') ()
+    push bs
+        | S.null bs' = NeedInput push close
+        | otherwise  = Done (Just bs') ()
+      where
+        bs' = S.dropWhile p bs
     close = return ()
 
 -- | Take the given number of bytes, if available.
 --
--- Since 0.2.0
-take :: Resource m => Int -> Sink S.ByteString m L.ByteString
-take n = L.fromChunks `liftM` (isolate n =$ CL.consume)
+-- Since 0.3.0
+take :: Monad m => Int -> Sink S.ByteString m L.ByteString
+take n0 =
+    go n0 id
+  where
+    go n front =
+        NeedInput push close
+      where
+        push bs =
+            case S.length bs `compare` n of
+                LT -> go (n - S.length bs) (front . (bs:))
+                EQ -> Done Nothing $ L.fromChunks $ front [bs]
+                GT ->
+                    let (x, y) = S.splitAt n bs
+                     in Done (Just y) $ L.fromChunks $ front [x]
+        close = return $ L.fromChunks $ front []
 
 -- | Split the input bytes into lines. In other words, split on the LF byte
 -- (10), and strip it from the output.
 --
--- Since 0.2.0
-lines :: Resource m => Conduit S.ByteString m S.ByteString
-lines =
-    conduitState id push close
+-- Since 0.3.0
+lines :: Monad m => Conduit S.ByteString m S.ByteString
+lines = NeedInput (push id) (close S.empty)
   where
-    push front bs' = return $ StateProducing leftover ls
+    push :: Monad m => (S.ByteString -> S.ByteString)
+                    -> S.ByteString
+                    -> Conduit S.ByteString m S.ByteString
+    push sofar more =
+        case S.uncons second of
+            Just (_, second') -> HaveOutput (push id second') (return ()) (sofar first)
+            Nothing ->
+                let rest = sofar more
+                 in NeedInput (push $ S.append rest) (close rest)
       where
-        bs = front bs'
-        (leftover, ls) = getLines id bs
+        (first, second) = S.breakByte 10 more
 
-    getLines front bs
-        | S.null bs = (id, front [])
-        | S.null y = (S.append x, front [])
-        | otherwise = getLines (front . (x:)) (S.drop 1 y)
-      where
-        (x, y) = S.breakByte 10 bs
+    close rest
+        | S.null rest = Done Nothing ()
+        | otherwise   = HaveOutput (Done Nothing ()) (return ()) rest
 
-    close front
-        | S.null bs = return []
-        | otherwise = return [bs]
-      where
-        bs = front S.empty
