@@ -21,6 +21,7 @@ module Data.Conduit.Text
     , utf32_be
     , ascii
     , iso8859_1
+    , lines
 
     ) where
 
@@ -29,7 +30,6 @@ import           Prelude hiding (head, drop, takeWhile, lines, zip, zip3, zipWit
 
 import           Control.Arrow (first)
 import qualified Control.Exception as Exc
-import           Control.Monad.Trans.Class (lift)
 import           Data.Bits ((.&.), (.|.), shiftL)
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
@@ -43,11 +43,12 @@ import           Data.Typeable (Typeable)
 
 import qualified Data.Conduit as C
 import qualified Data.Conduit.List as CL
-import Control.Monad.Trans.Resource (ResourceThrow (..))
+import Control.Monad.Trans.Resource (MonadThrow (..))
+import Control.Monad.Trans.Class (lift)
 
 -- | A specific character encoding.
 --
--- Since 0.2.0
+-- Since 0.3.0
 data Codec = Codec
     { codecName :: T.Text
     , codecEncode
@@ -64,58 +65,68 @@ instance Show Codec where
     showsPrec d c = showParen (d > 10) $
         showString "Codec " . shows (codecName c)
 
+-- | Emit each line separately
+--
+-- Since 0.4.1
+lines :: Monad m => C.Conduit T.Text m T.Text
+lines =
+    C.NeedInput (push id) (close T.empty)
+  where
+    push sofar more =
+        case T.uncons second of
+            Just (_, second') -> C.HaveOutput (push id second') (return ()) (sofar first')
+            Nothing ->
+                let rest = sofar more
+                 in C.NeedInput (push $ T.append rest) (close rest)
+      where
+        (first', second) = T.break (== '\n') more
+
+    close rest
+        | T.null rest = C.Done Nothing ()
+        | otherwise   = C.HaveOutput (C.Done Nothing ()) (return ()) rest
+
 -- | Convert text into bytes, using the provided codec. If the codec is
 -- not capable of representing an input character, an exception will be thrown.
 --
--- Since 0.2.0
-encode :: ResourceThrow m => Codec -> C.Conduit T.Text m B.ByteString
+-- Since 0.3.0
+encode :: MonadThrow m => Codec -> C.Conduit T.Text m B.ByteString
 encode codec = CL.mapM $ \t -> do
     let (bs, mexc) = codecEncode codec t
-    maybe (return bs) (resourceThrow . fst) mexc
+    maybe (return bs) (monadThrow . fst) mexc
 
 
 -- | Convert bytes into text, using the provided codec. If the codec is
 -- not capable of decoding an input byte sequence, an exception will be thrown.
 --
--- Since 0.2.0
-decode :: ResourceThrow m => Codec -> C.Conduit B.ByteString m T.Text
-decode codec = C.conduitState
-    Nothing
-    push
-    close
+-- Since 0.3.0
+decode :: MonadThrow m => Codec -> C.Conduit B.ByteString m T.Text
+decode codec =
+    C.NeedInput push (close B.empty)
   where
-    push mb input = do
-        (mb', ts) <- go' mb input
-        return $ C.StateProducing mb' ts
-    close mb =
-        case mb of
-            Nothing -> return []
-            Just b
-                | B.null b -> error "Data.Conduit.Text.decode: Received a null chunk"
-                | otherwise -> lift $ resourceThrow $ DecodeException codec (B.head b)
-
-    go' mb input = do -- FIXME This can be simplified significantly since input is now only a single BS
-        let bss = maybe id (:) mb [input]
-        either (lift . resourceThrow) return $ go bss id
-
-    go [] front = Right (Nothing, front [])
-    go (x:xs) front
-        | B.null x = go xs front
-    go (x:xs) front =
+    push bs =
         case extra of
-            Left (exc, _) -> Left exc
-            Right bs
-                | B.null bs -> go xs front'
-                | otherwise ->
-                    case xs of
-                        y:ys -> go (B.append bs y:ys) front'
-                        [] -> Right (Just bs, front' [])
+            Left (exc, _) -> C.PipeM (monadThrow exc) (monadThrow exc)
+            Right bs' ->
+                let push' = if B.null bs' then push else push . B.append bs'
+                    close' = close bs'
+                 in C.HaveOutput (C.NeedInput push' close') (close2 bs') text
       where
-        (text, extra) = codecDecode codec x
-        front' = front . (text:)
+        (text, extra) = codecDecode codec bs
+
+    close bs =
+        case B.uncons bs of
+            Nothing -> C.Done Nothing ()
+            Just (w, _) ->
+                let exc = monadThrow $ DecodeException codec w
+                 in C.PipeM exc (lift exc)
+
+    close2 bs =
+        case B.uncons bs of
+            Nothing -> return ()
+            Just (w, _) -> monadThrow $ DecodeException codec w
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 data TextException = DecodeException Codec Word8
                    | EncodeException Codec Char
     deriving (Show, Typeable)
@@ -148,7 +159,7 @@ splitSlowly dec bytes = valid where
             Right _ -> Right B.empty)
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 utf8 :: Codec
 utf8 = Codec name enc dec where
     name = T.pack "UTF-8"
@@ -181,7 +192,7 @@ utf8 = Codec name enc dec where
                     else decodeMore
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 utf16_le :: Codec
 utf16_le = Codec name enc dec where
     name = T.pack "UTF-16-LE"
@@ -208,7 +219,7 @@ utf16_le = Codec name enc dec where
         decodeAll = (TE.decodeUtf16LE bytes, B.empty)
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 utf16_be :: Codec
 utf16_be = Codec name enc dec where
     name = T.pack "UTF-16-BE"
@@ -243,7 +254,7 @@ utf16Required x0 x1 = required where
     x = (fromIntegral x1 `shiftL` 8) .|. fromIntegral x0
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 utf32_le :: Codec
 utf32_le = Codec name enc dec where
     name = T.pack "UTF-32-LE"
@@ -253,7 +264,7 @@ utf32_le = Codec name enc dec where
         Nothing -> splitSlowly TE.decodeUtf32LE bs
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 utf32_be :: Codec
 utf32_be = Codec name enc dec where
     name = T.pack "UTF-32-BE"
@@ -276,7 +287,7 @@ utf32SplitBytes dec bytes = split where
         else B.splitAt lenToDecode bytes
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 ascii :: Codec
 ascii = Codec name enc dec where
     name = T.pack "ASCII"
@@ -295,7 +306,7 @@ ascii = Codec name enc dec where
             else Left (DecodeException ascii (B.head unsafe), unsafe)
 
 -- |
--- Since 0.2.0
+-- Since 0.3.0
 iso8859_1 :: Codec
 iso8859_1 = Codec name enc dec where
     name = T.pack "ISO-8859-1"
