@@ -11,6 +11,7 @@ import qualified Data.Conduit.Lazy as CLazy
 import qualified Data.Conduit.Binary as CB
 import qualified Data.Conduit.Text as CT
 import Data.Conduit (runResourceT)
+import Data.Maybe (fromMaybe)
 import qualified Data.List as DL
 import Control.Monad.ST (runST)
 import Data.Monoid
@@ -22,10 +23,11 @@ import Data.ByteString.Lazy.Char8 ()
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Encoding as TLE
-import Control.Monad.Trans.Resource (runExceptionT_, withIO, resourceForkIO)
+import Control.Monad.Trans.Resource (runExceptionT_, allocate, resourceForkIO)
 import Control.Concurrent (threadDelay, killThread)
 import Control.Monad.IO.Class (liftIO)
 import Control.Applicative (pure, (<$>), (<*>))
+import Data.Functor.Identity (runIdentity)
 
 main :: IO ()
 main = hspecX $ do
@@ -52,7 +54,7 @@ main = hspecX $ do
     describe "ResourceT" $ do
         it "resourceForkIO" $ do
             counter <- I.newIORef 0
-            let w = withIO
+            let w = allocate
                         (I.atomicModifyIORef counter $ \i ->
                             (i + 1, ()))
                         (const $ I.atomicModifyIORef counter $ \i ->
@@ -78,9 +80,18 @@ main = hspecX $ do
             x <- runResourceT $ CL.sourceList [1..10] C.$$ CL.fold (+) (0 :: Int)
             x @?= sum [1..10]
         prop "is idempotent" $ \list ->
-            (runST $ runResourceT $ CL.sourceList list C.$$ CL.fold (+) (0 :: Int))
+            (runST $ CL.sourceList list C.$$ CL.fold (+) (0 :: Int))
             == sum list
 
+    describe "unfold" $ do
+        it "works" $ do
+            let f 0 = Nothing
+                f i = Just (show i, i - 1)
+                seed = 10 :: Int
+            x <- CL.unfold f seed C.$$ CL.consume
+            let y = DL.unfoldr f seed
+            x @?= y
+
     describe "Monoid instance for Source" $ do
         it "mappend" $ do
             x <- runResourceT $ (CL.sourceList [1..5 :: Int] `mappend` CL.sourceList [6..10]) C.$$ CL.fold (+) 0
@@ -120,6 +131,22 @@ main = hspecX $ do
             bs1 @=? bs2
             bs1 @=? bs3
 
+    describe "zipping" $ do
+        it "zipping two small lists" $ do
+            res <- runResourceT $ CL.zip (CL.sourceList [1..10]) (CL.sourceList [11..12]) C.$$ CL.consume
+            res @=? zip [1..10 :: Int] [11..12 :: Int]
+
+    describe "zipping sinks" $ do
+        it "take all" $ do
+            res <- runResourceT $ CL.sourceList [1..10] C.$$ CL.zipSinks CL.consume CL.consume
+            res @=? ([1..10 :: Int], [1..10 :: Int])
+        it "take fewer on left" $ do
+            res <- runResourceT $ CL.sourceList [1..10] C.$$ CL.zipSinks (CL.take 4) CL.consume
+            res @=? ([1..4 :: Int], [1..10 :: Int])
+        it "take fewer on right" $ do
+            res <- runResourceT $ CL.sourceList [1..10] C.$$ CL.zipSinks CL.consume (CL.take 4)
+            res @=? ([1..10 :: Int], [1..4 :: Int])
+
     describe "Monad instance for Sink" $ do
         it "binding" $ do
             x <- runResourceT $ CL.sourceList [1..10] C.$$ do
@@ -136,11 +163,10 @@ main = hspecX $ do
     describe "resumable sources" $ do
         it "simple" $ do
             (x, y, z) <- runResourceT $ do
-                bs <- C.bufferSource $ CL.sourceList [1..10 :: Int]
-                x <- bs C.$$ CL.take 5
-                y <- bs C.$$ CL.fold (+) 0
-                z <- bs C.$$ CL.consume
-                C.bsourceClose bs
+                let src1 = CL.sourceList [1..10 :: Int]
+                (src2, x) <- src1 C.$$+ CL.take 5
+                (src3, y) <- src2 C.$$+ CL.fold (+) 0
+                z <- src3 C.$$ CL.consume
                 return (x, y, z)
             x @?= [1..5] :: IO ()
             y @?= sum [6..10]
@@ -191,12 +217,12 @@ main = hspecX $ do
     describe "isolate" $ do
         it "bound to resumable source" $ do
             (x, y) <- runResourceT $ do
-                bsrc <- C.bufferSource $ CL.sourceList [1..10 :: Int]
-                x <- bsrc C.$= CL.isolate 5 C.$$ CL.consume
-                y <- bsrc C.$$ CL.consume
+                let src1 = CL.sourceList [1..10 :: Int]
+                (src2, x) <- src1 C.$= CL.isolate 5 C.$$+ CL.consume
+                y <- src2 C.$$ CL.consume
                 return (x, y)
             x @?= [1..5]
-            y @?= [6..10]
+            y @?= []
 
         it "bound to sink, non-resumable" $ do
             (x, y) <- runResourceT $ do
@@ -209,9 +235,9 @@ main = hspecX $ do
 
         it "bound to sink, resumable" $ do
             (x, y) <- runResourceT $ do
-                bsrc <- C.bufferSource $ CL.sourceList [1..10 :: Int]
-                x <- bsrc C.$$ CL.isolate 5 C.=$ CL.consume
-                y <- bsrc C.$$ CL.consume
+                let src1 = CL.sourceList [1..10 :: Int]
+                (src2, x) <- src1 C.$$+ CL.isolate 5 C.=$ CL.consume
+                y <- src2 C.$$ CL.consume
                 return (x, y)
             x @?= [1..5]
             y @?= [6..10]
@@ -242,6 +268,38 @@ main = hspecX $ do
             nums <- CLazy.lazyConsume $ mconcat $ map incr [1..10]
             liftIO $ nums @?= [1..10]
 
+        it' "returns nothing outside ResourceT" $ do
+            bss <- runResourceT $ CLazy.lazyConsume $ CB.sourceFile "test/main.hs"
+            bss @?= []
+
+    describe "sequence" $ do
+        it "simple sink" $ do
+            let sumSink :: C.MonadResource m => C.Sink Int m Int
+                sumSink = do
+                    ma <- CL.head
+                    case ma of
+                        Nothing -> return 0
+                        Just a  -> (+a) . fromMaybe 0 <$> CL.head
+
+            res <- runResourceT $ CL.sourceList [1..11]
+                             C.$= C.sequence sumSink
+                             C.$$ CL.consume
+            res @?= [3, 7, 11, 15, 19, 11]
+
+        it "sink with unpull behaviour" $ do
+            let sumSink :: C.MonadResource m => C.Sink Int m Int
+                sumSink = do
+                    ma <- CL.head
+                    case ma of
+                        Nothing -> return 0
+                        Just a  -> (+a) . fromMaybe 0 <$> CL.peek
+
+            res <- runResourceT $ CL.sourceList [1..11]
+                             C.$= C.sequence sumSink
+                             C.$$ CL.consume
+            res @?= [3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 11]
+
+
     describe "sequenceSink" $ do
         it "simple sink" $ do
             let sink () = do
@@ -282,19 +340,19 @@ main = hspecX $ do
 
     describe "text" $ do
         let go enc tenc cenc = do
-                prop (enc ++ " single chunk") $ \chars -> runST $ runExceptionT_ $ runResourceT $ do
+                prop (enc ++ " single chunk") $ \chars -> runST $ runExceptionT_ $ do
                     let tl = TL.pack chars
                         lbs = tenc tl
                         src = CL.sourceList $ L.toChunks lbs
                     ts <- src C.$= CT.decode cenc C.$$ CL.consume
                     return $ TL.fromChunks ts == tl
-                prop (enc ++ " many chunks") $ \chars -> runST $ runExceptionT_ $ runResourceT $ do
+                prop (enc ++ " many chunks") $ \chars -> runIdentity $ runExceptionT_ $ do
                     let tl = TL.pack chars
                         lbs = tenc tl
                         src = mconcat $ map (CL.sourceList . return . S.singleton) $ L.unpack lbs
                     ts <- src C.$= CT.decode cenc C.$$ CL.consume
                     return $ TL.fromChunks ts == tl
-                prop (enc ++ " encoding") $ \chars -> runST $ runExceptionT_ $ runResourceT $ do
+                prop (enc ++ " encoding") $ \chars -> runIdentity $ runExceptionT_ $ do
                     let tss = map T.pack chars
                         lbs = tenc $ TL.fromChunks tss
                         src = mconcat $ map (CL.sourceList . return) tss
@@ -306,6 +364,23 @@ main = hspecX $ do
         go "utf32_le" TLE.encodeUtf32LE CT.utf32_le
         go "utf32_be" TLE.encodeUtf32BE CT.utf32_be
 
+    describe "text lines" $ do
+        it "works across split lines" $
+            (CL.sourceList [T.pack "abc", T.pack "d\nef"] C.$= CT.lines C.$$ CL.consume) ==
+                [[T.pack "abcd", T.pack "ef"]]
+        it "works with multiple lines in an item" $
+            (CL.sourceList [T.pack "ab\ncd\ne"] C.$= CT.lines C.$$ CL.consume) ==
+                [[T.pack "ab", T.pack "cd", T.pack "e"]]
+        it "works with ending on a newline" $
+            (CL.sourceList [T.pack "ab\n"] C.$= CT.lines C.$$ CL.consume) ==
+                [[T.pack "ab"]]
+        it "works with ending a middle item on a newline" $
+            (CL.sourceList [T.pack "ab\n", T.pack "cd\ne"] C.$= CT.lines C.$$ CL.consume) ==
+                [[T.pack "ab", T.pack "cd", T.pack "e"]]
+        it "is not too eager" $ do
+            x <- CL.sourceList ["foobarbaz", error "ignore me"] C.$$ CT.decode CT.utf8 C.=$ CL.head
+            x @?= Just "foobarbaz"
+
     describe "binary isolate" $ do
         it "works" $ do
             bss <- runResourceT $ CL.sourceList (replicate 1000 "X")
@@ -315,12 +390,46 @@ main = hspecX $ do
     describe "unbuffering" $ do
         it "works" $ do
             x <- runResourceT $ do
-                bsrc <- C.bufferSource $ CL.sourceList [1..10 :: Int]
-                bsrc C.$$ CL.drop 5
-                let src = C.unbufferSource bsrc
-                src C.$$ CL.fold (+) 0
+                let src1 = CL.sourceList [1..10 :: Int]
+                (src2, ()) <- src1 C.$$+ CL.drop 5
+                src2 C.$$ CL.fold (+) 0
             x @?= sum [6..10]
 
+    describe "operators" $ do
+        it "only use =$=" $
+            runIdentity
+            (    CL.sourceList [1..10 :: Int]
+              C.$$ CL.map (+ 1)
+             C.=$= CL.map (subtract 1)
+             C.=$= CL.map (* 2)
+             C.=$= CL.map (`div` 2)
+             C.=$= CL.fold (+) 0
+            ) @?= sum [1..10]
+        it "only use =$" $
+            runIdentity
+            (    CL.sourceList [1..10 :: Int]
+              C.$$ CL.map (+ 1)
+              C.=$ CL.map (subtract 1)
+              C.=$ CL.map (* 2)
+              C.=$ CL.map (`div` 2)
+              C.=$ CL.fold (+) 0
+            ) @?= sum [1..10]
+        it "chain" $ do
+            x <-      CL.sourceList [1..10 :: Int]
+                C.$=  CL.map (+ 1)
+                C.=$= CL.map (+ 1)
+                C.$=  CL.map (+ 1)
+                C.=$= CL.map (subtract 3)
+                C.=$= CL.map (* 2)
+                C.$$  CL.map (`div` 2)
+                C.=$= CL.map (+ 1)
+                C.=$  CL.map (+ 1)
+                C.=$= CL.map (+ 1)
+                C.=$  CL.map (subtract 3)
+                C.=$= CL.fold (+) 0
+            x @?= sum [1..10]
+
+
     describe "properly using binary file reading" $ do
         it "sourceFile" $ do
             x <- runResourceT $ CB.sourceFile "test/random" C.$$ CL.consume
@@ -338,19 +447,19 @@ main = hspecX $ do
 
         prop "works" $ \bss' ->
             let bss = map S.pack bss'
-             in runST $ runResourceT $
+             in runIdentity $
                 CL.sourceList bss C.$$ go (L.fromChunks bss)
     describe "binary takeWhile" $ do
         prop "works" $ \bss' ->
             let bss = map S.pack bss'
-             in runST $ runResourceT $ do
+             in runIdentity $ do
                 bss2 <- CL.sourceList bss C.$$ CB.takeWhile (>= 5) C.=$ CL.consume
                 return $ L.fromChunks bss2 == L.takeWhile (>= 5) (L.fromChunks bss)
 
     describe "binary dropWhile" $ do
         prop "works" $ \bss' ->
             let bss = map S.pack bss'
-             in runST $ runResourceT $ do
+             in runIdentity $ do
                 bss2 <- CL.sourceList bss C.$$ do
                     CB.dropWhile (< 5)
                     CL.consume
@@ -393,15 +502,8 @@ main = hspecX $ do
                 src C.$= CB.isolate 10 C.$$ CL.head
             x @?= Just "foobarbazb"
 
-    describe "bufferedFuseLeft" $ do
-        it "does not double close conduit" $ do
-            x <- runResourceT $ do
-                bsrc <- C.bufferSource $ CL.sourceList ["foobarbazbin"]
-                bsrc C.$= CB.isolate 10 C.$$ CL.head
-            x @?= Just "foobarbazb"
-
     describe "binary" $ do
-        prop "lines" $ \bss' -> runST $ runResourceT $ do
+        prop "lines" $ \bss' -> runIdentity $ do
             let bss = map S.pack bss'
                 bs = S.concat bss
                 src = CL.sourceList bss
